package org.lukeallen;

import gnu.io.CommPort;
import gnu.io.CommPortIdentifier;
import gnu.io.SerialPort;
import gnu.io.SerialPortEvent;
import gnu.io.SerialPortEventListener;

import java.awt.EventQueue;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.Timer;
import java.util.TimerTask;

import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;
import javax.swing.JTextField;
import javax.swing.SpringLayout;

public class AX12Config {

	serialManager mySerialManager;
	Timer myTimer;
	int currentBaud;
	private JFrame frmAxConfigUtility;
	private JTextField servoIDTextField;
	private JTextField param1TextField;
	private JTextField param2NTextField;
	private JTextArea consoleTextArea;
	private JComboBox instructionComboBox;
	private JScrollPane outputScrollPane;
	private JButton btnSendWhileAutocycling;
	
	
	//******************************************************************
	//*	INSTRUCTIONS
	//******************************************************************
/*	#define iPING 		0x01 //obtain a status packet
	#define iREAD_DATA	0x02 //read Control Table values
	#define iWRITE_DATA	0x03 //write Control Table values
	#define iREG_WRITE 	0x04 //write and wait for ACTION instruction
	#define iACTION 	0x05 //triggers REG_WRITE instruction
	#define iRESET 		0x06 //set factory defaults
	#define iSYNC_WRITE 0x83 //simultaneously control multiple actuators*/
	
	

	/**
	 * Launch the application.
	 */
	public static void main(String[] args) {
		EventQueue.invokeLater(new Runnable() {
			public void run() {
				try {
					AX12Config window = new AX12Config();
					window.frmAxConfigUtility.setVisible(true);
				} catch (Exception e) {
					e.printStackTrace();
				}
			}
		});
	}

	/**
	 * Create the application.
	 */
	public AX12Config() {
		initialize();
		mySerialManager = new serialManager();
		try{
			mySerialManager.open("COM1", 115200);
			print("opened COM1");
		}catch(Exception e){
			print("Error opening serial port");
		}
	}
	
	/**
	 * Called by the window's event listener when the window is closed.
	 */
	private void closeApplication(){
		mySerialManager.close();
	}
	
	
	public void handleAllBaudButtonPress(){
		if(myTimer != null)
			myTimer.cancel();
		currentBaud = 0;
		myTimer = new Timer();
		myTimer.schedule(new TimerTask() {
            public void run() {
                baudloop();
            }
        }, 50,//delay before first run; doesn't really matter
                50);//repeat period in ms
	}
	
	private void baudloop(){
		//make a packet to tell my PIC code to set the desired BRG byte
		byte[] baudPacket = {(byte)0xFF,(byte)0xFF,(byte)255,(byte)3,(byte)3,(byte)currentBaud,(byte)0};
		addChecksumInLastSpot(baudPacket);
		print("Set baud: " + currentBaud);
		mySerialManager.xmit(baudPacket);
		try{
			Thread.sleep(25);//let PIC respond
		}catch(InterruptedException e){
		}
		//send the packet the user put in
		handleSendButtonPress();
		
		
		currentBaud++;
		if(currentBaud > 255)
			myTimer.cancel();
	}
	
	private void addChecksumInLastSpot(byte[] packet){
		int checksum=0;
		for(int i=2; i<packet.length-1;i++)
			checksum += ((int)packet[i] & 0xFF);
		packet[packet.length-1] = ((byte)(~checksum & 0xFF));//checksum is defined as sum with inverted bits
	}
	
	/**
	 * Send the packet. This gets called by the action listener for the button.
	 */
	public void handleSendButtonPress(){
		boolean valid = true;
		try{
			//get servo ID
			int servoID = Integer.parseInt(servoIDTextField.getText());
			if(servoID>255){
				valid = false;
				print("Invalid servo ID");
			}
			//get instruction
			int instruction = instructionComboBox.getSelectedIndex() + 1;
			if(instruction == 7)//SYNC WRITE instruction
				instruction = 83;
			//get address
			int param1 = Integer.parseInt(param1TextField.getText());
			if(param1 > 255){
				valid = false;
				print("invalid param1 (address) byte");
			}
			
			//determine param2 length. For READ_DATA, it's 1 byte
			int parambytes = 1;//instruction == 2
			if(instruction == 1 || instruction == 5 || instruction == 6){//PING, ACTION, and RESET
				parambytes = 0;
			}else if(instruction == 3 || instruction == 4){//write or REG_WRITE instruction
				if(		param1 == 6 ||
						param1 == 8 ||
						param1 == 14 ||
						param1 == 30 ||
						param1 == 32 ||
						param1 == 34 ||
						param1 == 48   ){
					parambytes = 3;        //determine proper data length in bytes (from table on PDF pg. 19 of AX-12 datasheet)
				}else{
					parambytes = 2;
				}
			}
			int param2 = Integer.parseInt(param2NTextField.getText());
			if((instruction == 3 || instruction == 4) && (param2 > Math.pow(2,8*(parambytes-1)) -1) ){
				valid = false;
				print("param2 is too large for the given instruction and address");
			}
			//packet holds 0xFF, 0xFF, id, total length-4, instruction, parameters, checksum 
			byte[] packet = new byte[5+parambytes+1];
			packet[0]=packet[1]=(byte)0xFF;
			packet[2]=(byte)servoID;
			packet[3]= (byte)(packet.length-4);
			packet[4]=(byte)instruction;
			if(parambytes > 0)
				packet[5] = (byte)param1;
			if(parambytes > 1)
				packet[6] = (byte)(param2 & 0xFF);//lower byte
			if(parambytes > 2)
				packet[7] = (byte)((param2 >> 8) & 0xFF);//upper byte
			int checksum=0;
			for(int i=2; i<packet.length-1;i++)
				checksum += ((int)packet[i] & 0xFF);
			packet[packet.length-1] = ((byte)(~checksum & 0xFF));//checksum is defined as sum with inverted bits
			if(valid){
				String xmission = "";
				for(int i=0; i<packet.length;i++)
					xmission += Integer.toString(((int)packet[i] & 0xFF)) + ",";
				print("transmitted: " + xmission);
				mySerialManager.xmit(packet);
			}
			else
				print("packet not transmitted; see error messages.");
			
		}
		catch(NumberFormatException e){
			print("Numbers must be written as decimals");
		}
	}
	
	void print(String message){
		consoleTextArea.append(message + "\n");
		//consoleTextArea.setText(consoleTextArea.getText() + message + "\n");//append
		consoleTextArea.setCaretPosition(consoleTextArea.getText().length());//scroll to last character added
	}

	/**
	 * Initialize the contents of the frame. This was mainly auto-generated using Google's WindowBuilder plugin to Eclipse
	 */
	private void initialize() {
		frmAxConfigUtility = new JFrame();
		frmAxConfigUtility.setTitle("AX12 Config Utility");
		frmAxConfigUtility.addWindowListener(new WindowAdapter() {
			@Override
			public void windowClosing(WindowEvent arg0) {
				closeApplication();
			}
		});
		frmAxConfigUtility.setBounds(100, 100, 608, 635);
		frmAxConfigUtility.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		SpringLayout springLayout = new SpringLayout();
		frmAxConfigUtility.getContentPane().setLayout(springLayout);
		
		servoIDTextField = new JTextField();
		springLayout.putConstraint(SpringLayout.NORTH, servoIDTextField, 10, SpringLayout.NORTH, frmAxConfigUtility.getContentPane());
		springLayout.putConstraint(SpringLayout.WEST, servoIDTextField, 10, SpringLayout.WEST, frmAxConfigUtility.getContentPane());
		springLayout.putConstraint(SpringLayout.EAST, servoIDTextField, 50, SpringLayout.WEST, frmAxConfigUtility.getContentPane());
		servoIDTextField.setText("1");
		frmAxConfigUtility.getContentPane().add(servoIDTextField);
		servoIDTextField.setColumns(10);
		
		JLabel lblId = new JLabel("Servo ID (1 is factory default, 254 means broadcast to all)");
		springLayout.putConstraint(SpringLayout.WEST, lblId, 6, SpringLayout.EAST, servoIDTextField);
		springLayout.putConstraint(SpringLayout.EAST, lblId, -50, SpringLayout.EAST, frmAxConfigUtility.getContentPane());
		frmAxConfigUtility.getContentPane().add(lblId);
		
		instructionComboBox = new JComboBox();
		springLayout.putConstraint(SpringLayout.WEST, instructionComboBox, 10, SpringLayout.WEST, frmAxConfigUtility.getContentPane());
		springLayout.putConstraint(SpringLayout.EAST, instructionComboBox, -157, SpringLayout.EAST, frmAxConfigUtility.getContentPane());
		springLayout.putConstraint(SpringLayout.SOUTH, lblId, -9, SpringLayout.NORTH, instructionComboBox);
		springLayout.putConstraint(SpringLayout.NORTH, instructionComboBox, 6, SpringLayout.SOUTH, servoIDTextField);
		instructionComboBox.addItem("PING 		0x01 obtain a status packet");
		instructionComboBox.addItem("READ_DATA	0x02 read Control Table values");
		instructionComboBox.addItem("WRITE_DATA	0x03 write Control Table values");
		instructionComboBox.addItem("REG_WRITE 	0x04 write and wait for ACTION instruction");
		instructionComboBox.addItem("ACTION 	0x05 triggers REG_WRITE instruction");
		instructionComboBox.addItem("RESET 		0x06 set factory defaults");
		instructionComboBox.addItem("SYNC_WRITE 0x83 simultaneously control multiple servos");
		frmAxConfigUtility.getContentPane().add(instructionComboBox);
		
		JLabel lblInstruction = new JLabel("Instruction");
		springLayout.putConstraint(SpringLayout.NORTH, lblInstruction, 4, SpringLayout.NORTH, instructionComboBox);
		springLayout.putConstraint(SpringLayout.WEST, lblInstruction, 6, SpringLayout.EAST, instructionComboBox);
		frmAxConfigUtility.getContentPane().add(lblInstruction);
		
		param1TextField = new JTextField();
		springLayout.putConstraint(SpringLayout.NORTH, param1TextField, 6, SpringLayout.SOUTH, instructionComboBox);
		springLayout.putConstraint(SpringLayout.WEST, param1TextField, 10, SpringLayout.WEST, frmAxConfigUtility.getContentPane());
		springLayout.putConstraint(SpringLayout.EAST, param1TextField, 41, SpringLayout.WEST, frmAxConfigUtility.getContentPane());
		param1TextField.setText("25");
		frmAxConfigUtility.getContentPane().add(param1TextField);
		param1TextField.setColumns(10);
		
		JLabel lblAddresswriteAs = new JLabel("Param1: Control table address (write as a decimal, not hex)");
		springLayout.putConstraint(SpringLayout.NORTH, lblAddresswriteAs, 3, SpringLayout.NORTH, param1TextField);
		springLayout.putConstraint(SpringLayout.WEST, lblAddresswriteAs, 6, SpringLayout.EAST, param1TextField);
		frmAxConfigUtility.getContentPane().add(lblAddresswriteAs);
		
		param2NTextField = new JTextField();
		param2NTextField.setText("0");
		springLayout.putConstraint(SpringLayout.NORTH, param2NTextField, 6, SpringLayout.SOUTH, param1TextField);
		springLayout.putConstraint(SpringLayout.WEST, param2NTextField, 0, SpringLayout.WEST, servoIDTextField);
		frmAxConfigUtility.getContentPane().add(param2NTextField);
		param2NTextField.setColumns(10);
		
		JLabel lblParameterValue = new JLabel("Parameter 2 thru n (write as a decimal).");
		springLayout.putConstraint(SpringLayout.NORTH, lblParameterValue, 3, SpringLayout.NORTH, param2NTextField);
		springLayout.putConstraint(SpringLayout.WEST, lblParameterValue, 6, SpringLayout.EAST, param2NTextField);
		frmAxConfigUtility.getContentPane().add(lblParameterValue);
		
		JButton sendButton = new JButton("Send Packet");
		sendButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent arg0) {
				handleSendButtonPress();
			}
		});
		springLayout.putConstraint(SpringLayout.NORTH, sendButton, 6, SpringLayout.SOUTH, param2NTextField);
		springLayout.putConstraint(SpringLayout.WEST, sendButton, 0, SpringLayout.WEST, servoIDTextField);
		frmAxConfigUtility.getContentPane().add(sendButton);
		
		outputScrollPane = new JScrollPane();
		JButton btnClear = new JButton("Clear");
		btnClear.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent arg0) {
				consoleTextArea.setText("");
			}
		});
		springLayout.putConstraint(SpringLayout.WEST, btnClear, 0, SpringLayout.WEST, lblInstruction);
		springLayout.putConstraint(SpringLayout.SOUTH, btnClear, 0, SpringLayout.SOUTH, outputScrollPane);
		frmAxConfigUtility.getContentPane().add(btnClear);
	
		springLayout.putConstraint(SpringLayout.NORTH, outputScrollPane, 23, SpringLayout.SOUTH, sendButton);
		springLayout.putConstraint(SpringLayout.WEST, outputScrollPane, 10, SpringLayout.WEST, frmAxConfigUtility.getContentPane());
		springLayout.putConstraint(SpringLayout.SOUTH, outputScrollPane, -10, SpringLayout.SOUTH, frmAxConfigUtility.getContentPane());
		springLayout.putConstraint(SpringLayout.EAST, outputScrollPane, 0, SpringLayout.EAST, instructionComboBox);
		frmAxConfigUtility.getContentPane().add(outputScrollPane);
		
		consoleTextArea = new JTextArea();
		consoleTextArea.setEditable(false);
		outputScrollPane.setViewportView(consoleTextArea);
		springLayout.putConstraint(SpringLayout.NORTH, consoleTextArea, -111, SpringLayout.SOUTH, frmAxConfigUtility.getContentPane());
		springLayout.putConstraint(SpringLayout.WEST, consoleTextArea, 265, SpringLayout.WEST, frmAxConfigUtility.getContentPane());
		springLayout.putConstraint(SpringLayout.SOUTH, consoleTextArea, -32, SpringLayout.SOUTH, frmAxConfigUtility.getContentPane());
		springLayout.putConstraint(SpringLayout.EAST, consoleTextArea, -10, SpringLayout.EAST, frmAxConfigUtility.getContentPane());
		
		btnSendWhileAutocycling = new JButton("Send while auto-cycling baud (requires my custom PIC code)");
		springLayout.putConstraint(SpringLayout.NORTH, btnSendWhileAutocycling, 0, SpringLayout.NORTH, sendButton);
		springLayout.putConstraint(SpringLayout.WEST, btnSendWhileAutocycling, 41, SpringLayout.EAST, sendButton);
		btnSendWhileAutocycling.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent arg0) {
				handleAllBaudButtonPress();
			}
		});
		frmAxConfigUtility.getContentPane().add(btnSendWhileAutocycling);
	}
	
	
	
	
	
	
	
	
	/*Holder of standard serial code, more or less copied from RxTx "Event Based Two Way Communication" example.
	 * Supports connection to one serial port at a time.
	 * Usage:
	 * Create an instance of it like this:
	 
	    //in your class's variable declarations:
 		serialManager mySerialManager;
 
 		//in your class's constructor, or when you want to open the port:
		mySerialManager = new serialManager();
		try{
			mySerialManager.open("COM1", 115200);
			print("opened COM1");
		}catch(Exception e){
			print("Error opening serial port");
		}
	
		//on application exit, or when you want to close the port:	 
		if(mySerialManager != null)
	 		mySerialManager.close()	
	 	
	 */
	class serialManager{
		SerialPort serialPort;
		InputStream inStream;
        OutputStream outStream;
        Thread xmitter;
		/* Pass this a port name, like "COM1"
		 * 
		 * 
		 */
		void open ( String portName , int baud) throws Exception
		   {
		       CommPortIdentifier portIdentifier = CommPortIdentifier.getPortIdentifier(portName);
		       if ( portIdentifier.isCurrentlyOwned() )
		       {
		           System.out.println("Error: Port is currently in use");
		       }
		       else
		       {
		           CommPort commPort = portIdentifier.open(this.getClass().getName(),2000);
		           
		           if ( commPort instanceof SerialPort )
		           {
		               serialPort = (SerialPort) commPort;
		               serialPort.setSerialPortParams(baud,//baud rate. 115200 is the fastest PIC supports, 9600 is the slowest.
												       SerialPort.DATABITS_8,
												       SerialPort.STOPBITS_1,
												       SerialPort.PARITY_NONE);
		               
		               inStream = serialPort.getInputStream();
		               outStream = serialPort.getOutputStream();

		               
		               serialPort.addEventListener(new SerialReader(inStream));
		               serialPort.notifyOnDataAvailable(true);
		           }
		           else
		           {
		               System.out.println("Error: Only serial ports are handled by this example.");
		           }
		       }     
		   }
		void close(){
			try{
			if(serialPort != null)
				serialPort.removeEventListener();
			}catch(Exception e){
				
			}
		}
		
		boolean xmit(byte[] xmission){
			try{
				for(int i=0; i<xmission.length; i++){
					outStream.write(xmission[i]);//the outputStream is buffered, so there's no need to do this in a thread or anything like that.
	            }  
			}catch(IOException e){
				return false;//return false if error
			}
			return true;
		}
		
		void listPorts()
	    {
			@SuppressWarnings("all")
	        java.util.Enumeration<CommPortIdentifier> portEnum = CommPortIdentifier.getPortIdentifiers();
	        while ( portEnum.hasMoreElements() ) 
	        {
	            CommPortIdentifier portIdentifier = portEnum.nextElement();
	            print(portIdentifier.getName()  +  " - " +  getPortTypeName(portIdentifier.getPortType()) );
	        }        
	    }
	    
	    String getPortTypeName ( int portType )
	    {
	        switch ( portType )
	        {
	            case CommPortIdentifier.PORT_I2C:
	                return "I2C";
	            case CommPortIdentifier.PORT_PARALLEL:
	                return "Parallel";
	            case CommPortIdentifier.PORT_RAW:
	                return "Raw";
	            case CommPortIdentifier.PORT_RS485:
	                return "RS485";
	            case CommPortIdentifier.PORT_SERIAL:
	                return "Serial";
	            default:
	                return "unknown type";
	        }
	    }
		
	    
	    
	    /**
	     * Handles the input coming from the serial port. A new line character
	     * is treated as the end of a block in this example. 
	     */
	    class SerialReader implements SerialPortEventListener 
	    {
	        private InputStream in;
	        private byte[] buffer = new byte[1024];
	        public SerialReader ( InputStream in )
	        {
	            this.in = in;
	        }
	        public void serialEvent(SerialPortEvent arg0) {
	            int data;
	            try
	            {
	                int len = 0;
	                while ( ( data = in.read()) > -1 )//stop and print if you reach end of buffer
	                {
	                    /*if ( data == '\n' ) {
	                        break;
	                    }*/
	                    buffer[len++] = (byte) data;
	                    print(Integer.toString(data));//this is my custom print function
	                }
	                System.out.print(new String(buffer,0,len));
	                
	            }
	            catch ( IOException e )
	            {
	                e.printStackTrace();
	                System.exit(-1);
	            }             
	        }
	    }
	
	    /*
	    class SerialWriter implements Runnable 
	    {
	        OutputStream out;
	        byte[] myxmission;
	        public SerialWriter ( OutputStream out, byte[] xmission )
	        {
	        	myxmission = xmission;
	            this.out = out;
	        }
	        public void run ()
	        {
	            try
	            {                
	                for(int i=0; i<myxmission.length; i++){
	                    this.out.write(myxmission[i]);
	                }                
	            }
	            catch ( IOException e )
	            {
	                e.printStackTrace();
	                System.exit(-1);
	            }            
	        }
	    }
	    */
	}

	
}






 